use std::borrow::Cow;

use crate::router::AppState;
use crate::zitadel::{IntrospectionConfig, IntrospectionState};

use async_sqlx_session::PostgresSessionStore;
use axum::extract::{FromRef, FromRequestParts, Query, State};
use axum::headers::authorization::Basic;
use axum::headers::{Authorization, Cookie};
use axum::http::request::Parts;
use axum::http::StatusCode;
use axum::http::{header::SET_COOKIE, HeaderMap};
use axum::response::{IntoResponse, Redirect};
use axum::routing::get;
use axum::{async_trait, RequestPartsExt, Router, TypedHeader};
use axum_sessions::async_session::{Session, SessionStore};
use como_domain::users::User;
use como_infrastructure::register::ServiceRegister;
use oauth2::basic::BasicClient;
use oauth2::{reqwest::async_http_client, AuthorizationCode, CsrfToken, Scope, TokenResponse};
use oauth2::{RedirectUrl, TokenIntrospectionResponse};
use serde::Deserialize;
use zitadel::oidc::introspection::introspect;

#[derive(Debug, Deserialize)]
pub struct ZitadelAuthParams {
    return_url: Option<String>,
}

pub async fn zitadel_auth(State(client): State<BasicClient>) -> impl IntoResponse {
    let (auth_url, _csrf_token) = client
        .authorize_url(CsrfToken::new_random)
        .add_scope(Scope::new("identify".to_string()))
        .add_scope(Scope::new("openid".to_string()))
        .url();

    Redirect::to(auth_url.as_ref())
}

pub static COOKIE_NAME: &str = "SESSION";

#[derive(Debug, Deserialize)]
#[allow(dead_code)]
pub struct AuthRequest {
    code: String,
    state: String,
}

pub async fn login_authorized(
    Query(query): Query<AuthRequest>,
    State(store): State<PostgresSessionStore>,
    State(oauth_client): State<BasicClient>,
    State(introspection_state): State<IntrospectionState>,
) -> impl IntoResponse {
    let token = oauth_client
        .exchange_code(AuthorizationCode::new(query.code.clone()))
        .request_async(async_http_client)
        .await
        .unwrap();

    let config = IntrospectionConfig::from_ref(&introspection_state);

    let res = introspect(
        &config.introspection_uri,
        &config.authority,
        &config.authentication,
        token.access_token().secret(),
    )
    .await
    .unwrap();

    let mut session = Session::new();
    session
        .insert(
            "user",
            User {
                id: res.sub().unwrap().into(),
            },
        )
        .unwrap();

    let cookie = store.store_session(session).await.unwrap().unwrap();

    let cookie = format!("{}={}; SameSite=Lax; Path=/", COOKIE_NAME, cookie);

    let mut headers = HeaderMap::new();
    headers.insert(SET_COOKIE, cookie.parse().unwrap());

    (headers, Redirect::to("http://localhost:3000/dash/home"))
}

pub struct AuthController;

impl AuthController {
    pub async fn new_router(
        _service_register: ServiceRegister,
        app_state: AppState,
    ) -> anyhow::Result<Router> {
        Ok(Router::new()
            .route("/zitadel", get(zitadel_auth))
            .route("/authorized", get(login_authorized))
            .with_state(app_state))
    }
}

pub struct UserFromSession {
    pub user: User,
}

#[async_trait]
impl<S> FromRequestParts<S> for UserFromSession
where
    PostgresSessionStore: FromRef<S>,
    BasicClient: FromRef<S>,
    IntrospectionState: FromRef<S>,
    S: Send + Sync,
{
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let store = PostgresSessionStore::from_ref(state);

        let cookie: Option<TypedHeader<Cookie>> = parts.extract().await.unwrap();

        let session_cookie = cookie.as_ref().and_then(|cookie| cookie.get(COOKIE_NAME));
        if let None = session_cookie {
            let introspection_state = IntrospectionState::from_ref(state);

            let basic: Option<TypedHeader<Authorization<Basic>>> = parts.extract().await.unwrap();

            if let Some(basic) = basic {
                let config = IntrospectionConfig::from_ref(&introspection_state);

                let res = introspect(
                    &config.introspection_uri,
                    &config.authority,
                    &config.authentication,
                    basic.password(),
                )
                .await
                .unwrap();

                return Ok(UserFromSession {
                    user: User {
                        id: res.sub().unwrap().into(),
                    },
                });
            }

            return Err((StatusCode::UNAUTHORIZED, "No session was found"));
        }

        let session_cookie = session_cookie.unwrap();

        tracing::debug!(
            "UserFromSession: got session cookie from user agent, {}={}",
            COOKIE_NAME,
            session_cookie
        );
        // continue to decode the session cookie
        let user =
            if let Some(session) = store.load_session(session_cookie.to_owned()).await.unwrap() {
                if let Some(user) = session.get::<User>("user") {
                    tracing::debug!(
                        "UserFromSession: session decoded success, user_id={:?}",
                        user.id
                    );
                    user
                } else {
                    return Err((
                        StatusCode::INTERNAL_SERVER_ERROR,
                        "No `user_id` found in session",
                    ));
                }
            } else {
                tracing::debug!(
                    "UserIdFromSession: err session not exists in store, {}={}",
                    COOKIE_NAME,
                    session_cookie
                );
                return Err((StatusCode::BAD_REQUEST, "No session found for cookie"));
            };

        Ok(UserFromSession { user })
    }
}
